import os, imp, re, sys, logging

from identification import class_identifier \
                         , default_validity \
                         , no_validity 

from interface.base.verify import IVerifiable
from interface.user import *

class class_registry(object):
    """
    Class registry so that objects can be loaded dynamically.
    """
    NO_CHECK = -1
    DEFAULT_CHECK = 0

    VALIDITY_CHECK = {}
    VALIDITY_CHECK[NO_CHECK] = no_validity
    VALIDITY_CHECK[DEFAULT_CHECK] = default_validity

    def __init__(self, validity_check_type, log):
        self.logger = log
        self.loaded_classes = {}
        self.loaded_modules = {}
        self.validity_check = self.VALIDITY_CHECK[validity_check_type]
    
    def generate_class_identifiers(self, path, classes):
        log = self.logger
        coll = []
        log.debug("Generating class identifier collection")
        for class_identifier_string in classes.split():
            log.debug("parsing: %s" % (class_identifier_string))
            try:
                name, version_tuple_str = class_identifier_string.split("-")
                major, minor, revision = [int(i) for i in version_tuple_str.split(".")]
                coll.append(class_identifier(path, name, major, minor, revision, self.validity_check))
            except Exception, ex:
                log.error(ex)
        return coll

    def load_classes_from_module(self, path, classes):
        """
        loads the classes from a specific module
        """

        log = self.logger
        loaded_modules = self.loaded_modules
        loaded_classes = self.loaded_classes

        clsids = self.generate_class_identifiers(path, classes)
        log.debug("Reviewing %d class identifier(s)." % (len(clsids)))

        PYTHON_FILE_EXTENSION_PATTERN = "py$"
        is_python_module = re.compile(PYTHON_FILE_EXTENSION_PATTERN, re.I)

        module = None
        module_name = None

        for clsid in clsids:

            if not os.path.exists(clsid.path):
                raise IOError("File does not exist:  %s" % (clsid.path))

            match = is_python_module.search(clsid.path)
            if not match:
                raise Exception("Must be a valid python module with file extension:  py\n\nFile Path:  %s\nFile Pattern: %s" % (clsid.path,PYTHON_FILE_EXTENSION_PATTERN))
            
            # determine what the module name is using the basename.
            module_name = os.path.basename(clsid.path).split(".py")[0]

            if module_name in loaded_modules.keys():
                log.info("Module %s is already loaded." % (clsid.path))            
            else:
                try:
                    log.debug("loading %s from %s" % (module_name, clsid.path))
                    loaded_modules[module_name] = imp.load_source(module_name, clsid.path)
                except:
                    log.error(sys.exc_info())
                    continue

            module = loaded_modules[module_name]

            # load the classes from the module according to what was requested.        
            if clsid.name in loaded_classes.keys():
                log.error("%s cannot be loaded twice." % (name))
                continue

            try:
                cls_factory = getattr(module, clsid.name)
                if clsid.is_valid(clsid, cls_factory, log):
                    loaded_classes[clsid.__name__] = (clsid, cls_factory)
            except:
                log.error(sys.exc_info())

    def instantiate(self, class_name, *args, **kwargs):

        log = self.logger
        loaded_classes = self.loaded_classes
        if class_name not in loaded_classes.keys():
            log.debug(loaded_classes.keys())
            log.error("%s is not in the list of loaded classes:  %s" % (class_name, loaded_classes.keys()))
            return None

        clsid, cls_factory = loaded_classes[class_name]
        inst = cls_factory(*args,**kwargs)
        clsid.is_valid(clsid, inst, log)
        return inst





from zope.interface import Attribute
from potatoe.interface.base.verify import IVerifiable

class ITextFileAccessor(IVerifiable):
    """
    An object which is able to provide both the path to 
    a file and the ability to get the contents of this 
    file.
    """
    path = Attribute("""The path that the text file exists on""")
    contents = Attribute("""The contents of the text file.""")
    
    
    def contents():
        """
        Returns the contents of the file.
        """
    def open(file_access_details):
        """
        open the connection to the file.
        """
    def close():
        """
        close the connection to the file.
        """
    def __enter__(file_access_details):
        """
        Opens the file and keeps it open. 
        """
    def __exit__():
        """
        Closes and destroys the connection to the file.
        """

class ICommaSeperatedValueFileAccessor(ITextFileAccessor):
    """
    """

    __fields__ = Attribute("The list of fields available in this file.")

    def fields():
        """
        The list of fields available in this file.
        """
    def to_dictionary_list():
        """
        Returns a list of dictionaries where the keys are the field 
        names and the values are the actual values in that row. 
        """
    def to_tuples():
        """
        Returns a tuple of tuples of all the data.
        """
    
class IDataAccessor(IVerifiable):
    """
    An object which is able to provide access to data.
    """

    __conn__ = Attribute("The connection being used to access the data.")
    __conn_details__ = Attribute("The details used to create the connection.")

    def get(request, *args , **kwargs ):
        """
        Gets an object based on the request.
        """
    def update(request, *args , **kwargs ):
        """
        Updates the object based on the request.
        """
    def insert(request, *args , **kwargs ):
        """
        Inserts the object based on the request.
        """
    def delete(request, *args , **kwargs ):
        """
        Deletes the object based on the request.
        """
    def connect(connection_details, *args , **kwargs ):
        """
        Connects to the data store.
        """
    def disconnect():
        """
        Disconnects from the data store.
        """
    def __enter__(connection_details):
        """
        Creates the connection to the data store and opens it.
        """
    def __exit__():
        """
        Closes and destroys the connection to the data store.
        """




"""
This module is used to identify classes.  It includes
methods that can be used to validate that the class
in question meets the expectations of the system.
"""

import logging, sys, traceback, re

from zope.interface.interface import InterfaceClass
from zope.interface.verify import verifyObject, verifyClass
from interface.base.verify import IVerifiable

def no_validity(self, cls, log):
    return True

def default_validity(clsid, cls, log, **kwargs):
    try:
        name = cls.__name__
        log.info("Performing the default validity check on the class factory %s" % (name))

        is_valid_version = version_check(cls, clsid)
        if not is_valid_version:
            return False

        is_implements_stated_interfaces = interface_check(cls)
        if not is_implements_stated_interfaces:
            return False

        log.info("%s passed the default validity check!" % (name))

        return True 

    except Exception, ex:
        log.error(ex)
        return False

def version_check(cls, clsid, log = logging.getLogger()):
    if len(cls.__version__) != 3:
        raise Exception("Version unavailable for comparison in dynamic class detail instance.")
    if len(cls.__name__) == 0:
        raise Exception("Name unavailable.")
    major, minor, revision = cls.__version__
    min_major, min_minor, min_revision = clsid.__version__
    if major < min_major:
        if minor < min_minor:
            if revision < min_revision:
                log.debug("%s Version Check Failed:  %s.%s.%s < %s.%s.%s" % (clsid.__name__,major,minor,revision,min_major,min_minor,min_revision))
                return False
    return True

def verify_iface(iface,cls,log):

    if not isinstance(iface,InterfaceClass):
        raise Exception("%s object must be of type %s" % (type(iface), InterfaceClass))

    log.info("Verifying %s on %s" % (iface,cls))
    if callable(cls):
        try:
            verifyClass(iface,cls)
            return True
        except:
            log.error("%s factory does not implement %s" % (cls,iface))
            log.error(sys.exc_info())
            raise
    else:    
        try:
            verifyObject(iface,cls)
            return True
        except:
            log.error("%s instance does not implement %s" % (cls,iface))
            log.error(sys.exc_info())
            raise

    return False 

def interface_check(cls, log = logging.getLogger()):
    """
    Confirms the class matches the new class matches the expected interface.
    """
    if not verify_iface(IVerifiable,cls,log):
        return False
    try:
        for iface in cls.__interface__:
            if not isinstance(iface,InterfaceClass):
                raise TypeError()
            if not verify_iface(iface,cls,log):
                return False
    except TypeError:
        iface = cls.__interface__
        if not isinstance(iface,InterfaceClass):
            log.error("%s %s stated interface must be of type %s" % (iface, type(iface),InterfaceClass))
            return False
        if not verify_iface(iface,cls,log):
            return False
    return True

class class_identifier(object):
    def __init__(self, path, name, maj, min, rev, validity_check_callable = no_validity, log = logging.getLogger()):
        self.log = log
        self.__name__ = name
        self.__version__ = (maj, min, rev)
        self.__path__ = path
        if not callable(validity_check_callable):
            raise Exception("Validity Check method parameter must be callable.")
        self.is_valid = validity_check_callable
    @property
    def path(self):
        return self.__path__
    @property
    def name(self):
        return self.__name__
    def __str__(self):
        return "<class_identifier, name=%s, version=%s, path=%s, is_valid=%s>" \
            % (   self.__name__         \
                , str(self.__version__) \
                , self.__path__         \
                , str(self.is_valid.__name__))





from zope.interface import Interface, Attribute

class IVerifiable(Interface):
    """
    An object which provides an interface. 
    """
    __interface__ = Attribute("The interface which the object defines.")
    __name__ = Attribute("The name of the object.")
    __version__ = Attribute("The version of the object.")



import os
from setuptools import setup, find_packages

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "potato",
    version = "0.0.2",
    author = "Sean Ochoa",
    author_email = "sean.m.ochoa@gmail.com",
    description = ("a dynamic class loader"),
    license = "Eclipse Public License -v 1.0",
    keywords = "dynamic class loading",
    url = "https://github.com/seanochoa/potato",
    packages=find_packages(exclude=["test"]),
    long_description=read('README'),
    classifiers = [
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Intended Audience :: Developers",
        "License :: OSI Approved",
        "Operating System :: OS Independent"
    ],
    install_requires=['zope.interface']
)
